#!/usr/bin/python3
# -*- coding: utf-8 -*-
# USAGE: MSG4.py YYYYmmDDHHMM, where DD=01..31, HH=00..23, MM=00,15,30,45
#************************************************************************
#
# European geostationary satellite Meteosat-11 (MSG4) at 0.0 degrees East
#------------------------------------------------------------------------
# EUMETSAT is EUMETCast disseminating MSG4 data in EUMETSAT_Data_Channel_2
# Meteosat-11 (MSG4), at 0.0 noon is exactly 12:00 UTC, instrument 'seviri'
#
# Channel | VIS006 |   HRV  | VIS008 | IR_016 | IR_039 | WV_062 |
# WaveLen |  0.635 |   0.7  |   0.81 |   1.64 |   3.90 |   6.25 |
# --------+--------+--------+--------+--------+--------+--------+
# Channel | WV_073 | IR_087 | IR_097 | IR_108 | IR_120 | IR_134 |
# WaveLen |   7.35 |   8.70 |   9.66 |  10.80 |  12.00 |  13.40 |
#
# CH-3123 Belp, 2021/01/17       License GPL3          (c) Ernst Lobsiger
#
#************************************************************************

# I need
import os
import sys
import shutil

# Pytroll/SatPy needs these
from satpy import Scene
from glob import glob

# Needed for grid labels that have their own color setting (see above grid definition)
# This did segfault for a while and has then been fixed by Dave (may or may not work)!
# import aggdraw

# Why to hell is it not working?
# from satpy.utils import debug_on
# debug_on()

# Minimal command line parameter test
usage = sys.argv[0] + ' expects YYYYmmDDHHMM as CLI parameter:\n' \
      + 'See comment concerning EUMETCast in the script header.'
if len(sys.argv) != 2:
    sys.exit(usage)
if len(sys.argv[1]) != 12:
    sys.exit(usage)

# Time slot from CLI
Dat=sys.argv[1]

# Get substrings
Yea=Dat[:4]
Mon=Dat[4:6]
Day=Dat[6:8]
Hou=Dat[8:10]
Min=Dat[10:]


# vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
# ********** ADAPT THESE PARAMETERS TO YOUR PERSONAL NEEDS ***********

# Edit 1-3 parameter(s) below according to your file layout:
# My files are in a /MountPoint/Channel/YYYY/mm/DD structure
# No trailing / because this will make timestamps unreadable
segdir = '/srv/rec_0/Channel_2' + '/' + Yea + '/' + Mon + '/' + Day
# Windows toodrp = 'C:', datdrp = 'D:' / GNU/Linux use homedir
toodrp = '/home/eumetcast'
datdrp = '/home/eumetcast'


# Available composites in this script (far more in ../satpy/etc/composites/seviri.yaml):
# 'overview', 'natural_color', 'realistic_colors', 'hrv_clouds', 'night_fog', 'ir108_3d'
composite = 'overview'


# Choose your area (remember that HRV data has only limited coverage)
# area = 'seviri_0deg'
# area = 'eurol'
area = 'westminster'
# area = 'isleofman'
# area = 'switzerland'


# Configure your individual overlays, either True/False or 1/0 do work
ADDgrid    =  True
ADDcoasts  =  True
ADDborders =  True
ADDrivers  =  True
ADDpoints  =  True

# OVRCache speeds up static overlays. These are stored in your .../EMCdata/cache
# With caching you will sometimes have to delete *.png OVRcache files manually !
OVRcache   =  False
# GEOcache allows for a resample cache of GEO satellites (speedy massproduction)
# These caches are stored in subdirectories with unique names .../nn_lut-*.zarr
GEOcache   =  False

# ******** TOUCH THE CODE BELOW ONLY IF YOU KNOW WHAT YOU DO *********
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^


# Set temporary directory and image directory
tmpdir = datdrp + '/EMCdata/tmpdirs/xmsg4'
imgdir = datdrp + '/EMCdata/images/Meteosat-11'

# Dir must exist
try:
   os.chdir(tmpdir)
except:
   sys.exit('Cannot change to tmp directory ...')

# Clean tmp directory
os.system('rm -f H-000*')

# Output of global_scene.available_composite_names() when all seviri channels are loaded
# **************************************************************************************
# ['airmass', 'ash', 'cloudtop', 'cloudtop_daytime', 'colorized_ir_clouds', 'convection',
# 'day_microphysics', 'day_microphysics_winter', 'dust', 'fog', 'green_snow', 'hrv_clouds',
# 'hrv_fog', 'hrv_severe_storms', 'hrv_severe_storms_masked', 'ir108_3d', 'ir_cloud_day',
# 'ir_overview', 'ir_sandwich', 'natural_color', 'natural_color_nocorr', 'natural_color_raw',
# 'natural_color_with_night_ir', 'natural_color_with_night_ir_hires', 'natural_enh',
# 'natural_with_night_fog', 'night_fog', 'night_ir_alpha', 'night_ir_with_background',
# 'night_ir_with_background_hires', 'night_microphysics', 'overview', 'overview_raw',
# 'realistic_colors', 'snow', 'vis_sharpened_ir']

# We only decompress what is actually needed, see file seviri.yaml
needed_files = {'overview'         : ['VIS006', 'VIS008', 'IR_108'],
                'natural_color'    : ['VIS006', 'VIS008', 'IR_016'],
                'realistic_colors' : ['VIS006', 'VIS008', 'HRV'],
                'hrv_clouds'       : ['IR_108', 'HRV'],
                'night_fog'        : ['IR_039', 'IR_108', 'IR_120', 'IR_134'],
                'ir108_3d'         : ['IR_108']}
# Some composite names are too long for IM annotation left of pic
abbreviation = {'overview'         : 'overview',
                'natural_color'    : 'natural_color',
                'realistic_colors' : 'realist_color',
                'hrv_clouds'       : 'hrv_clouds',
                'night_fog'        : 'night_fog',
                'ir108_3d'         : 'ir108_3d'}

for b in needed_files[composite]:
    files = glob(segdir + '/H-000-MSG4*' + str(b) + '*' + Dat + '*-C_')
    # Decompress all files found
    for f in files:
        os.system(toodrp + '/EMCtools/exefiles/xRITDecompress ' + f)

# Meteosat-11 sends only 1 PROlogue and 1 EPIlogue file per 15 minute time slot
# But e.g. the hrit_goes reader insisted that there is a separate PRO per band?
files = glob(segdir + '/H-000-MSG4*PRO*' + Dat + '*')
for f in files:
    shutil.copy2(f, tmpdir);
files = glob(segdir + '/H-000-MSG4*EPI*' + Dat + '*')
for f in files:
    shutil.copy2(f, tmpdir);

# Now all the files needed should be decompressed in tmpdir
files = glob(tmpdir + '/H-000-MSG4*' + Dat + '*')

# We should have got at least some good files
if files == []:
    sys.exit('Sorry, no good files found ...')

global_scene = Scene(filenames = files, reader = 'seviri_l1b_hrit')
global_scene.load([composite])


# This is a speedup by projection/area cache. It works for GEOstationary satellites only. 
if GEOcache:
    geocache = datdrp + '/EMCdata/cache/MSG4'
else:
    geocache = None

# Full earth seviri image has South on top, image size is 3712x3712 pixels, resample rotates it (no need for IM)
new_scene = global_scene.resample(area, radius_of_influence = 20000, cache_dir = geocache) #, reduce_data = False)

# Next is line by Christian Peters, thanks a lot!
# try:except: added after Graham Woolf's problems
try: # This is for real (multichannel) composites
    layers, height, width = new_scene[composite].shape
except: # In case of a single channel 'composite'
    height, width = new_scene[composite].shape

# Autoscale font_size (fs) of grid labels
fs = int((height + width) / 100)

# my_font = aggdraw.Font('red', datdrp + '/EMCdata/fonts/DejaVuSerif.ttf',
#           opacity = 255, size = fs) # aggdraw.Font takes only 4 arguments
# Font color can either be specified 'white' or (R, G, B) = (255, 255, 255)
# .. 'font': my_font} below may or may not work anymore now or in the future.
# If a direct font definition does not work, use the inline definition below:

# New interface to pycoast + POIs have been integrated 2020, thanks Martin !!
my_grid  = {'major_lonlat': (10,10), 'minor_lonlat': (2, 2),
            # Opacity 0 will hide the line, values 0 ... 255  EL
            'outline': (255, 255, 255) , 'outline_opacity': 255,
            'minor_outline': (200, 200, 200),'minor_outline_opacity': 127,
            'width': 1.5, 'minor_width': 1.0, 'minor_is_tick': False,
            'write_text': True, 'lat_placement': 'lr', 'lon_placement': 'b',
            'font': datdrp + '/EMCdata/fonts/DejaVuSerif.ttf', 'font_size': fs}
            # minor_is_tick: False draws a line not ticks, True does ticks
            # label placement l,r,lr for latitude and t,b,tb for longitude

# Colors are specified as strings (140 HTML color names) or as triple (R, G, B) 8Bit.
# Best 'resolution' is automatically choosen according to area_def, no need to touch.
# Coasts also has keys ..,'fill': 'green', 'fill_opacity': 50} for land mass filling.
# 'level' defaults to 1 but should be choosen at least 2 for rivers (1 = 0 in GSHHG).
my_coasts  = {'outline': (255, 255,   0), 'width': 1.5, 'level': 1} # , 'resolution': 'i'}  # level 1 .. 4
my_borders = {'outline': (255,   0,   0), 'width': 1.0, 'level': 1} # , 'resolution': 'h'}  # level 1 .. 3
my_rivers  = {'outline': (  0,   0, 255), 'width': 1.0, 'level': 3} # , 'resolution': 'f'}  # level 1 .. 11

# List of POIs (Station, Cities, etc. ...)
my_poi_list = [((13.4052, 52.5214), 'Berlin'),
               (( 2.3522, 48.8566), 'Paris'),
               (( 0.1278, 51.5074), 'London'),
               ((-3.1880, 55.9539), 'Edinbourgh'),
               ((-6.2598, 53.3514), 'Dublin'),
               (( 7.4989, 46.8918), 'Belp')]

# You may find other True Type Fonts (*.ttf) in /usr/share/fonts/truetype/...
# Windows 10 PRO users may find more True Type Fonts in C:\Windows\Fonts\ ...
my_points = {'font': datdrp + '/EMCdata/fonts/DejaVuSerif.ttf',
             'font_size': 20, 'points_list': my_poi_list, 'symbol': 'circle',
             'ptsize': 20, 'outline': 'black', 'fill': 'red', 'width': 5.0,
             'outline_opacity': 255,'fill_opacity': 255, 'box_outline': 'navy',
             'box_linewidth': 3.0, 'box_fill': 'gold', 'box_opacity': 255}

# Cache is useful for speedup. Overlays are stored as tranparent *.png per satellite!
my_cache = {'file': datdrp + '/EMCdata/cache/MSG4/' + composite, 'regenerate': False}

# Setup indivdual overlays dictionary
my_overlays = {}
if ADDgrid:    # Add key grid
    my_overlays['grid'] = my_grid
if ADDcoasts:  # Add key coasts
    my_overlays['coasts'] = my_coasts
if ADDborders: # Add key borders
    my_overlays['borders'] = my_borders
if ADDrivers:  # Add key rivers
    my_overlays['rivers'] = my_rivers
if ADDpoints:  # Add key points
    my_overlays['points'] = my_points
if OVRcache:   # Add key cache
    my_overlays['cache'] = my_cache

# Enhance image with overlays and save *.png to temporary directory
new_scene.save_dataset(composite, 'MSG4.png',
          overlay = {'coast_dir': datdrp + '/EMCdata/gshhg-shp',
         'overlays' : my_overlays}) # 2^6=32 combinations should do

# ****************************************************************
# NEW POST PROCESSING: WORKS WITH ALL SCRIPTS AND PROJECTION AREAS
# ****************************************************************

# ImageMagick (IM) postprocessing, added directory logos in EMCdata
# logo1 = datdrp + '/EMCdata/logos/GNU-Linux-Logo-Penguin-SVG.png'
logo1 = datdrp + '/EMCdata/logos/eumetsat_200x199.png'
logo2 = datdrp + '/EMCdata/logos/PyTROLL_400x400.jpg'

# Use IM to annotate (maybe also -equalize) and convert to *.jpg
# www.imagemagick.org/Usage/crop/#splice or .../annotating/#labeling
# This is the syntax used -draw "text OffsetX,OffsetY 'Hello World!'"
# This annotates to the left, numbers depend on original image size
# Colors can be {dark,light}blue, etc. or rgb(R,G,B) with \( escaped

def magick_string(EffHeight):
    """
    It seems ImageMagick under Windows cannot cope with multiline text
    that includes \n newlines. We have to draw every line separately.
    This scales all pixel values for IM according to the image height.
    Should make it easier to use annotation and logos at the left side.
    This has been optimized using MSG4 full scan with 3712x3712 pixels.
    """
    OptHeight = 3712
    OptSplice = 640
    OptLogo1s = 540
    OptLogo1x = 50   # 40 for 'GNU-Linux-Logo-Penguin-SVG.png'
    OptLogo1y = 40   # 60 for 'GNU-Linux-Logo-Penguin-SVG.png'
    OptLogo2s = 530
    OptLogo2x = 55
    OptLogo2y = 55
    OptPoints = 90

    fac = EffHeight/OptHeight
    EffSplice = str(int(fac*OptSplice+0.5))
    EffLogo1s = str(int(fac*OptLogo1s+0.5))
    EffLogo1x = str(int(fac*OptLogo1x+0.5))
    EffLogo1y = str(int(fac*OptLogo1y+0.5))
    EffLogo2s = str(int(fac*OptLogo2s+0.5))
    EffLogo2x = str(int(fac*OptLogo2x+0.5))
    EffLogo2y = str(int(fac*OptLogo2y+0.5))
    EffPoints = str(int(fac*OptPoints+0.5))

    legend = [' \'DATE:\'\" '     , ' \''+Yea+'/'+Mon+'/'+Day+'\'\" ',
              ' \'TIME:\'\" '     , ' \''+Hou+':'+Min+' UTC'+'\'\" ',
              ' \'SOURCE:\'\" '   , ' \'EUMETCast\'\" ',
              ' \'SERVICE:\'\" '  , ' \'Basic\'\" ',
              ' \'CHANNEL:\'\" '  , ' \'Channel_2\'\" ',
              ' \'RECEIVER:\'\" ' , ' \'TBS-6909X\'\" ',
              ' \'SATELLITE:\'\" ', ' \'Meteosat-11\'\" ',
              ' \'SENSOR:\'\" '   , ' \'seviri\'\" ',
              ' \'COMPOSITE:\'\" ', ' \''+abbreviation[composite]+'\'\" ',
              ' \'AREA:\'\" '     , ' \''+area+'\'\" ']
    multidraw = ''
    yoff = fac*470
    for n in range(0, len(legend)):
        if (n % 2) == 0:
            xoff = fac*40
            yoff += fac*150
        else:
            xoff = fac*60
            yoff += fac*100
        multidraw += '-draw \"text +'+str(int(xoff+0.5))+'+'+str(int(yoff+0.5))+legend[n]

    cmdstr = 'convert MSG4.png '\
    + '-gravity west -background rgb\(245,245,245\) -splice '+EffSplice+'x0 '\
    + logo1+' -gravity northwest -geometry '+EffLogo1s+'x'\
    + EffLogo1s+'+'+EffLogo1x+'+'+EffLogo1y+' -composite '\
    + logo2+' -gravity southwest -geometry '+EffLogo2s+'x'\
    + EffLogo2s+'+'+EffLogo2x+'+'+EffLogo2y+' -composite '\
    + '-pointsize '+EffPoints+' -gravity northwest '+multidraw\
    + imgdir+'/'+'Meteosat-11-'+Dat[:8]+'-SLO-'+Dat[8:]+'-'+composite+'-'+area+'.jpg'

    return(cmdstr)

# *Image Magick*
os.system(magick_string(height))
# IM will still be working when this python script returns (subprocess.call?)
